import logging

from django.db import models

from .mixins import TenantManagerMixin, TenantModelMixin, TenantQuerySet

logger = logging.getLogger(__name__)


class TenantManager(TenantManagerMixin, models.Manager.from_queryset(TenantQuerySet)):
    # This parameter is required to add Tenenat Manager in models in the migrations.
    use_in_migrations = True


class TenantModel(TenantModelMixin, models.Model):
    # Abstract model which all the models related to tenant inherit.

    objects = TenantManager()

    class Meta:
        abstract = True



from django.conf import settings
from django.test.signals import setting_changed
from django.utils.module_loading import import_string

SETTINGS_NAMESPACE = "ISLANDS"

USER_SETTINGS = getattr(settings, SETTINGS_NAMESPACE, None)

DEFAULTS = {
    "ALLOW_NULL_TENANT": False,
    "ALLOW_CHANGE_TENANT": False,
    "USE_ASGIREF": False,
}

IMPORT_STRINGS = []

REMOVED_SETTINGS = []


def perform_import(val, setting_name):
    if val is None:
        return None

    if isinstance(val, str):
        return import_from_string(val, setting_name)

    if isinstance(val, list | tuple):
        return [import_from_string(item, setting_name) for item in val]

    return val


def import_from_string(val, setting_name):
    try:
        return import_string(val)
    except ImportError as e:
        msg = f"Could not import '{val}' for API setting '{setting_name}'. {e.__class__.__name__}: {e}."
        raise ImportError(msg) from e


class AppSettings:
    def __init__(self, user_settings=None, defaults=None, import_strings=None):
        if user_settings:
            self._user_settings = self.__check_user_settings(user_settings)

        self.defaults = defaults or DEFAULTS
        self.import_strings = import_strings or IMPORT_STRINGS
        self._cached_attrs = set()

    @property
    def user_settings(self):
        if not hasattr(self, "_user_settings"):
            self._user_settings = getattr(settings, SETTINGS_NAMESPACE, {})
        return self._user_settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError(f"Invalid API setting: '{attr}'")

        try:
            val = self.user_settings[attr]
        except KeyError:
            val = self.defaults[attr]

        if attr in self.import_strings:
            val = perform_import(val, attr)

        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    def __check_user_settings(self, user_settings):
        for setting in REMOVED_SETTINGS:
            if setting in user_settings:
                msg = f"The '{setting}' setting has been removed. Please refer to documentation for available settings."
                raise RuntimeError(msg)
        return user_settings

    def reload(self):
        for attr in self._cached_attrs:
            delattr(self, attr)

        self._cached_attrs.clear()

        if hasattr(self, "_user_settings"):
            delattr(self, "_user_settings")


app_settings = AppSettings(USER_SETTINGS, DEFAULTS, IMPORT_STRINGS)


def reload_api_settings(*args, **kwargs):
    setting = kwargs["setting"]
    if setting == SETTINGS_NAMESPACE:
        app_settings.reload()


setting_changed.connect(reload_api_settings)



import inspect

from django.apps import apps

from .settings import app_settings

if app_settings.USE_ASGIREF:
    # asgiref must be installed, its included with Django >= 3.0
    from asgiref.local import Local as local
else:
    from threading import local


_thread_locals = _context = local()


def get_model_by_db_table(db_table):
    for model in apps.get_models():
        if model._meta.db_table == db_table:
            return model

    raise ValueError(f"No model found with db_table {db_table}")


def get_current_tenant():
    return getattr(_context, "tenant", None)


def get_tenant_column(model_class_or_instance: object):
    if inspect.isclass(model_class_or_instance):
        model_class_or_instance = model_class_or_instance()

    try:
        return model_class_or_instance.tenant_field
    except Exception as not_a_tenant_model:
        raise ValueError(
            f"{model_class_or_instance.__class__.__name__} is not an instance or a subclass of TenantModel or does not inherit from TenantMixin"
        ) from not_a_tenant_model


def get_tenant_field(model_class_or_instance: object):
    tenant_column = get_tenant_column(model_class_or_instance)
    all_fields = model_class_or_instance._meta.fields
    try:
        return next(field for field in all_fields if field.column == tenant_column)
    except StopIteration as no_field_found:
        raise ValueError(
            f'No field found in {model_class_or_instance.__class__.__name__} with column name "{tenant_column}"'
        ) from no_field_found


def get_object_tenant(instance):
    field = get_tenant_field(instance)

    if field.primary_key:
        return instance

    return getattr(instance, field.name, None)


def set_object_tenant(instance, value):
    if instance.tenant_value is None and value and not isinstance(value, list):
        setattr(instance, instance.tenant_field, value)


def get_current_tenant_value():
    current_tenant = get_current_tenant()
    if not current_tenant:
        return None

    try:
        current_tenant = list(current_tenant)
    except TypeError:
        return current_tenant.tenant_value

    values = []
    for t in current_tenant:
        values.append(t.tenant_value)
    return values


def get_tenant_filters(table, filters=None):
    filters = filters or {}

    current_tenant_value = get_current_tenant_value()

    if not current_tenant_value:
        return filters

    if isinstance(current_tenant_value, list):
        filters[f"{get_tenant_column(table)}__in"] = current_tenant_value
    else:
        filters[get_tenant_column(table)] = current_tenant_value

    return filters


def set_current_tenant(tenant):
    setattr(_context, "tenant", tenant)


def unset_current_tenant():
    setattr(_context, "tenant", None)



import django

# default_app_config is auto detected for versions 3.2 and higher:
# https://docs.djangoproject.com/en/3.2/ref/applications/#for-application-authors
if django.VERSION < (3, 2):
    default_app_config = "django_islands.apps.DjangoIslands"

__all__ = ["default_app_config", "version"]

version = (0, 0, 2)

__version__ = ".".join(map(str, version))
__author__ = "Davi Silva Rafacho"
__copyright__ = "Copyright 2025, Yellow Devs"
__license__ = "MIT"
__maintainer__ = "Davi Silva Rafacho"
__email__ = "davi.s.rafacho.developer@gmail.com"
__status__ = "Production"
